#include <stdlib.h>
#include <ctype.h>
#include <math.h>
#include <stdio.h>
#include <sys/time.h>
#include <assert.h>
#include <string.h>

/* Optionally include OpenMP with the -fopenmp flag */
#if defined(_OPENMP)
    #include <omp.h>
#endif

#include "include/lbfgs.h"
#include "include/twister.h"

#include "include/plm.h"
#include "include/inference.h"

/* Numerical bounds for ZeroAPCPriors */
#define LAMBDA_J_MIN 1E-2
#define LAMBDA_J_MAX 1E4

/* Internal to InferPairModel: 
   Bayesian estimation of hyperparameters for sites by MCMC (HMC) */
void EstimateSiteParamsBayes(numeric_t *lambdas, alignment_t *ali, 
    options_t *options);

/* Internal to EstimateSiteParamsBayes: Hamiltonian Monte Carlo */
typedef numeric_t (*hmc_hfun_t) (void *params, const numeric_t *x, const int n);
typedef void (*hmc_gradfun_t) (void *params, const numeric_t *x, numeric_t *g,
    const int n);
numeric_t HamiltonianMonteCarlo(hmc_hfun_t hfun, hmc_gradfun_t grad, 
    void *params, numeric_t *X, int n, int s, int L, numeric_t *epsilon,
    int warmup);
numeric_t RandomNormal();

/* Internal to EstimateSiteParamsBayes: Hierarchical model for single sites */
numeric_t HMCSiteH(void *params, const numeric_t *x, const int n);
void HMCSiteHGrad(void *params, const numeric_t *x, numeric_t *g, const int n);

/* Internal to EstimateSiteParamsBayes: Hierarchical reparameterization */
numeric_t HMCSiteHNonCenter(void *params, const numeric_t *x, const int n);
void HMCSiteHGradNonCenter(void *params, const numeric_t *x, numeric_t *g, 
    const int n);

/* Internal to EstimateSiteParamsBayes: Quantile estimation */
numeric_t QuickSelect(numeric_t *A, int len, int k);

/* Internal to InferPairModel: 
   Objective functions for point parameter estimates (MAP) */
static lbfgsfloatval_t PLMNegLogPosterior(void *instance,
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n,
    const lbfgsfloatval_t step);
static lbfgsfloatval_t PLMNegLogPosteriorGapReduce(void *instance,
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n,
    const lbfgsfloatval_t step);
static lbfgsfloatval_t PLMNegLogPosteriorBlock(void *instance,
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n,
    const lbfgsfloatval_t step);
static lbfgsfloatval_t PLMNegLogPosteriorDO(void *instance,
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n,
    const lbfgsfloatval_t step);

/* Internal to InferPairModel: progress reporting */
static int ReportProgresslBFGS(void *instance, const lbfgsfloatval_t *x,
    const lbfgsfloatval_t *g, const lbfgsfloatval_t fx,
    const lbfgsfloatval_t xnorm, const lbfgsfloatval_t gnorm,
    const lbfgsfloatval_t step, int n, int k, int ls);

/* Internal to InferPairModel: parameter processing */
void PreCondition(const lbfgsfloatval_t *x, lbfgsfloatval_t *g,
    alignment_t *ali, options_t *options);
void PostCondition(const lbfgsfloatval_t *x, lbfgsfloatval_t *g,
    alignment_t *ali, options_t *options);
void ZeroAPCPriors(alignment_t *ali, options_t *options, numeric_t *lambdas,
    lbfgsfloatval_t *x);

/* Internal to InferPairModel: utility functions to L-BFGS */
const char *LBFGSErrorString(int ret);


numeric_t *InferPairModel(alignment_t *ali, options_t *options) {
    /* Estimate the MAP parameters of a maximum entropy model for a
       multiple sequence alignment */

    /* Initialize the regularization parameters */
    numeric_t *lambdas =
    (numeric_t *) malloc((ali->nSites + ali->nSites * (ali->nSites - 1) / 2)
            * sizeof(numeric_t));
    for (int i = 0; i < ali->nSites; i++) lambdaHi(i) = options->lambdaH;
    for (int i = 0; i < ali->nSites - 1; i++)
        for (int j = i + 1; j < ali->nSites; j++)
            lambdaEij(i, j) = options->lambdaE;

    /* For gap-reduced problems, eliminate the gaps and reduce the alphabet */
    if (options->algorithm == INFER_PLM_GAPREDUCE) {
        ali->nCodes = strlen(ali->alphabet) - 1;
        for (int i = 0; i < ali->nSites; i++)
            for (int s = 0; s < ali->nSeqs; s++)
                seq(s, i) -= 1;
    }

    /* Initialize parameters */
    ali->nParams = ali->nSites * ali->nCodes
        + ali->nSites * (ali->nSites - 1) / 2 * ali->nCodes * ali->nCodes;
    lbfgsfloatval_t *x = lbfgs_malloc(ali->nParams);
    if (x == NULL) {
        fprintf(stderr,
            "ERROR: Failed to allocate a memory block for variables.\n");
        exit(1);
    }
    for (int i = 0; i < ali->nParams; i++) x[i] = 0.0;

    /* Initialize site parameters with the ML estimates 
        hi = log(fi) + C
        A single pseudocount is added for stability 
       (Laplace's rule or Morcos et al. with lambda = nCodes) */
    if (options->zeroAPC != 1) {
        numeric_t pseudoC = (numeric_t) ali->nCodes;
        numeric_t Zinv = 1.0 / (ali->nEff + pseudoC);
        for (int i = 0; i < ali->nSites; i++)
            for (int ai = 0; ai < ali->nSites; ai++)
                xHi(i, ai) = Zinv * pseudoC / (numeric_t) ali->nCodes;
        for (int s = 0; s < ali->nSeqs; s++)
            for (int i = 0; i < ali->nSites; i++)
                xHi(i, seq(s, i)) += ali->weights[s] * Zinv;
        for (int i = 0; i < ali->nSites; i++)
            for (int ai = 0; ai < ali->nCodes; ai++)
                xHi(i, ai) = log(xHi(i, ai));
        /* Zero-sum gauge */
        for (int i = 0; i < ali->nSites; i++) {
            numeric_t hSum = 0.0;
            for (int ai = 0; ai < ali->nCodes; ai++) hSum += xHi(i, ai);
            numeric_t hShift = hSum / (numeric_t) ali->nCodes;
            for (int ai = 0; ai < ali->nCodes; ai++)
                xHi(i, ai) -= hShift;
        }
    }

    /* Fully Bayesian estimation of site hyperparameters */
    if (options->bayesLH) EstimateSiteParamsBayes(lambdas, ali, options);

    /* Start timer */
    gettimeofday(&ali->start, NULL);

    /* Initialize L-BFGS */
    lbfgs_parameter_t param;
    lbfgs_parameter_init(&param);
    param.epsilon = 1E-3;
    param.max_iterations = options->maxIter; /* 0 is unbounded */

    /* Array of void pointers provides relevant data structures */
    void *d[3] = {(void *)ali, (void *)options, (void *)lambdas};

    /* Estimate parameters by optimization */
    static lbfgs_evaluate_t algo;
    switch(options->algorithm) {
        case INFER_PLM:
            algo = PLMNegLogPosterior;
            break;
        case INFER_PLM_GAPREDUCE:
            algo = PLMNegLogPosteriorGapReduce;
            break;
        case INFER_PLM_BLOCK:
            algo = PLMNegLogPosteriorBlock;
            break;
        case INFER_PLM_DROPOUT:
            algo = PLMNegLogPosteriorDO;
            break;
        default:
            algo = PLMNegLogPosterior;
    }

    if (options->zeroAPC == 1) fprintf(stderr,
            "Estimating coupling hyperparameters le = 1/2 inverse variance\n");

    int ret = 0;
    lbfgsfloatval_t fx;
    ret = lbfgs(ali->nParams, x, &fx, algo, ReportProgresslBFGS,
        (void*)d, &param);
    fprintf(stderr, "Gradient optimization: %s\n", LBFGSErrorString(ret));

    /* Optionally re-estimate parameters with adjusted hyperparameters */
    if (options->zeroAPC == 1) {
        /* Form new priors on the variances */
        ZeroAPCPriors(ali, options, lambdas, x);

        /* Reinitialize coupling parameters */
        for (int i = 0; i < ali->nSites - 1; i++)
            for (int j = i + 1; j < ali->nSites; j++)
                for (int ai = 0; ai < ali->nCodes; ai++)
                    for (int aj = 0; aj < ali->nCodes; aj++)
                        xEij(i, j, ai, aj) = 0.0;

        /* Iterate estimation with new hyperparameter estimates */
        options->zeroAPC = 2;
        ret = lbfgs(ali->nParams, x, &fx, algo,
            ReportProgresslBFGS, (void*)d, &param);
        fprintf(stderr, "Gradient optimization: %s\n", LBFGSErrorString(ret));
    }

    /* Restore the alignment encoding after inference */
    if (options->algorithm == INFER_PLM_GAPREDUCE) {
        for (int i = 0; i < ali->nSites; i++)
            for (int s = 0; s < ali->nSeqs; s++)
                seq(s, i) += 1;
    }

    return x;
}

void EstimateSiteParamsBayes(numeric_t *lambdas, alignment_t *ali,
    options_t *options) {
    /* Estimate site hyperparameters lambda_hi = 1 / (2 * sigma_hi^2) by HMC */
    /* Each site is modeled as a Boltzmann-like distribution where the weights 
       for each state are drawn from a Gaussian of unknown variance */

    fprintf(stderr, "Estimating site hyperparameters lh = 1/2 inverse variance\n");
    fprintf(stderr, "site\t<lh>\t95%% credible\teps\taccept rate\n");
    for (int i = 0; i < ali->nSites; i++) {
        numeric_t *C = (numeric_t *) malloc(ali->nCodes * sizeof(numeric_t));
        for (int ai = 0; ai < ali->nCodes; ai++) C[ai] = fi(i, ai) * ali->nEff;

        /* ------------------------------_DEBUG_----------------------------- */
        /* Rescale the counts */
        // for (int ai = 0; ai < ali->nCodes; ai++) C[ai] = (C[ai] + 10) * 100.0;
        // for (int ai = 0; ai < ali->nCodes; ai++) C[ai] *= 100.0;
        // for (int ai = 0; ai < ali->nCodes; ai++) C[ai] /= 10.0;
        // for (int ai = 0; ai < ali->nCodes; ai++) C[ai] *= 100.0 / ali->nEff;
        /* ------------------------------^DEBUG^----------------------------- */

        /* Array of void pointers provides relevant parameters */
        numeric_t scale = 2;
        numeric_t counts = 0;
        for (int ai = 0; ai < ali->nCodes; ai++) counts += C[ai];
        void *params[3] = {(void *)C, (void *)&scale, (void *)&counts};

        /* Tuned parameters for HMC */
        int n = ali->nCodes + 1;        /* Number of variables */
        int s = 1E3;                    /* Number of samples to collect */
        int L = 100;                    /* Integration number of steps */
        numeric_t eps = 0.01;           /* Median step size +/- 10-fold */
        int warmup = 1E3;               /* Number of samples to discard */

        /* Samples from HMC */
        numeric_t *X = (numeric_t *) malloc(n * s * sizeof(numeric_t));
        for (int ix = 0; ix < n * s; ix++) X[ix] = 0;
        numeric_t accRate =
            HamiltonianMonteCarlo(HMCSiteHNonCenter, HMCSiteHGradNonCenter,
            params, X, n, s, L, &eps, warmup);

        /* Estimate posterior average from HMC */
        numeric_t invS = 1.0 / (numeric_t) s;
        numeric_t sigmaAv = 0;
        for (int sx = 0; sx < s; sx++) sigmaAv += invS * exp(X[sx * n]);

        /* Compute statistics of lambda = 1/2 the inverse variance */
        numeric_t *lambdah = (numeric_t *) malloc(s * sizeof(numeric_t));
        for (int sx = 0; sx < s; sx++)
            lambdah[sx] = 1.0 / (2.0 * exp(2.0 * X[sx * n]));
        numeric_t lambdaAv = 0;
        numeric_t lambdaVar = 0;
        for (int sx = 0; sx < s; sx++) lambdaAv += invS * lambdah[sx];
        for (int sx = 0; sx < s; sx++)
            lambdaVar += invS * pow(lambdah[sx] - lambdaAv, 2);

        /* Find the 2.5% and 97.5% quantiles for the 95% credible interval */
        numeric_t lowLambda = QuickSelect(lambdah, s, (int) floor(0.025 * s));
        for (int sx = 0; sx < s; sx++)
            lambdah[sx] = 1.0 / (2.0 * exp(2.0 * X[sx * n]));
        numeric_t highLambda = QuickSelect(lambdah, s, (int) floor(0.975 * s));
        free(lambdah);

        fprintf(stderr, "%d\t%3.2f\t[%3.2f, %3.2f]\t%5.4f\t%3.1f\n", i + 1, 
            lambdaAv, lowLambda, highLambda, eps, 100 * accRate);

        /* Set the hyperparameters with the estimated posterior mean */
        lambdaHi(i) = lambdaAv;

        /* ------------------------------_DEBUG_----------------------------- */
        /* Compute and print posterior means */
        // for (int ix = 1; ix < n; ix++) {
        //     numeric_t avH = 0;
        //     for (int sx = 0; sx < s; sx++) 
        //         avH += exp(X[sx * n]) * X[sx * n + ix];
        //     avH /= (numeric_t) s;
        //     fprintf(stdout, "%f\t", avH);
        // }
        // fprintf(stdout, "\n");
        // if (i == ali->nSites - 1) exit(0);
        /* ------------------------------^DEBUG^----------------------------- */

        /* ------------------------------_DEBUG_----------------------------- */
        /* Dump HMC samples for analysis */
        // if (i == 42) {
        //     for (int sx = 0; sx < s; sx++) {
        //         for (int ix = 0; ix < n; ix++)
        //             fprintf(stdout, "%f\t", X[sx * n + ix]);
        //         fprintf(stdout, "\n");
        //     }
        //     exit(0);
        // }
        /* ------------------------------^DEBUG^----------------------------- */
        free(X);
        free(C);
    }

    /* --------------------------------_DEBUG_--------------------------------*/
    /* Print marginals to console */
    // for (int i = 0; i < ali->nSites; i++) {
    //     if (i%10 == 0) {
    //        for (int ai = 0; ai < ali->nCodes; ai++)
    //         fprintf(stderr, "\033[0;30;47m  %c   \x1B[0m",
    //             ali->alphabet[ai +
    //                 (options->algorithm == INFER_PLM_GAPREDUCE)]);
    //         fprintf(stderr, "\n"); 
    //     }
    //     for (int ai = 0; ai < ali->nCodes; ai++)
    //         if (fi(i,ai) > 0.5) {
    //             fprintf(stderr, "\033[0;30;47m%5.1f\x1B[0m ", 100 * fi(i,ai));
    //         } else if (fi(i,ai) > 1.0 / ((numeric_t) ali->nCodes)) {
    //             fprintf(stderr, "%5.1f ", 100 * fi(i,ai));
    //         } else {
    //             fprintf(stderr, "      ");
    //         }
    //     fprintf(stderr, "\t%4.3f\n", lambdaHi(i));
    // }
    // exit(0);
    /* --------------------------------^DEBUG^--------------------------------*/
}

numeric_t HamiltonianMonteCarlo(hmc_hfun_t hfun, hmc_gradfun_t grad,
    void *params, numeric_t *X, int n, int s, int L, numeric_t *epsilon,
    int warmup) {
    /* Initialize position(x), gradient(g), and momentum(p) */
    numeric_t *x = (numeric_t *) malloc(n * sizeof(numeric_t));
    numeric_t *xnew = (numeric_t *) malloc(n * sizeof(numeric_t));
    numeric_t *g = (numeric_t *) malloc(n * sizeof(numeric_t));
    numeric_t *gnew = (numeric_t *) malloc(n * sizeof(numeric_t));
    numeric_t *p = (numeric_t *) malloc(n * sizeof(numeric_t));
    for (int i = 0; i < n; i++) x[i] = 0;
    for (int i = 0; i < n; i++) xnew[i] = 0;
    for (int i = 0; i < n; i++) g[i] = 0;
    for (int i = 0; i < n; i++) gnew[i] = 0;
    for (int i = 0; i < n; i++) p[i] = 0;
    numeric_t U = hfun(params, x, n);
    grad(params, x, g, n);

    /* Use the Mersenne Twister for reproducible sampling results */
    init_genrand(42);

    /* Step sizes are log-normally distribution with median EPS */
    numeric_t eps = *epsilon;
    numeric_t epsMu = log(eps);
    numeric_t epsStd = log(4.0);


    /* Store running averages of the statistics for mass tuning */
    numeric_t *xMean = (numeric_t *) malloc(n * sizeof(numeric_t));
    numeric_t *xxMean = (numeric_t *) malloc(n * sizeof(numeric_t));
    for (int i = 0; i < n; i++) xMean[i] = 0;
    for (int i = 0; i < n; i++) xxMean[i] = 0;
    int xCount = 0;

    /* Mass matrix (diagonal) scales the step sizes along each dimension */
    numeric_t *invMass = (numeric_t *) malloc(n * sizeof(numeric_t));
    numeric_t *massStd = (numeric_t *) malloc(n * sizeof(numeric_t));
    for (int i = 0; i < n; i++) invMass[i] = 1.0;
    for (int i = 0; i < n; i++) massStd[i] = 1.0;

    /* --------------- Warmup steps to estimate the mass matrix ------------- */
    int massStart = warmup / 3;
    int massEnd = 2 * warmup / 3;
    int adaptStart = 0;
    int adaptEnd = warmup;
    int wsteps = 0;
    /* Lagged number of accepted moves over past 10 proposals */
    int acceptCount = 0;
    int windowCount = 0;
    do {
        /* Randomize momentum */
        for (int i = 0; i < n; i++) p[i] = RandomNormal() * massStd[i];
        numeric_t K = 0;
        for (int i = 0; i < n; i++) K += p[i] * p[i] * invMass[i];
        K *= 0.5;
        numeric_t H = U + K;

        /* Integrate system by L leapfrog steps */
        eps = exp(epsStd * RandomNormal() + epsMu);
        for (int i = 0; i < n; i++) xnew[i] = x[i];
        for (int i = 0; i < n; i++) gnew[i] = g[i];
        for (int lx = 0; lx < L; lx++) {
            for (int i = 0; i < n; i++) p[i] -= eps * gnew[i] * 0.5;
            for (int i = 0; i < n; i++) xnew[i] += eps * p[i] * invMass[i];
            grad(params, xnew, gnew, n);
            for (int i = 0; i < n; i++) p[i] -= eps * gnew[i] * 0.5;
        }

        /* Evaluate the proposal */
        numeric_t Unew = hfun(params, xnew, n);
        numeric_t Knew = 0;
        for (int i = 0; i < n; i++) Knew += p[i] * p[i] * invMass[i];
        Knew *= 0.5;
        numeric_t Hnew = Unew + Knew;
        numeric_t dH = Hnew - H;
        if (dH < 0 || ((numeric_t) genrand_real3()) < exp(-dH)) {
            /* Accept the proposal */
            U = Unew;
            for (int i = 0; i < n; i++) x[i] = xnew[i];
            for (int i = 0; i < n; i++) g[i] = gnew[i];
            wsteps++;
            acceptCount++;

            /* Store running averages of the statistics for mass tuning */
            if (wsteps >= massStart && wsteps <= massEnd) {
                for (int i = 0; i < n; i++) xMean[i] += x[i];
                for (int i = 0; i < n; i++) xxMean[i] += x[i] * x[i];
                xCount++;
            }

            /* Use the sample variance of each variable for inverse mass */
            if (wsteps == massEnd) {
                numeric_t invCounts = 1.0 / (numeric_t) xCount;
                for (int i = 0; i < n; i++) xMean[i] *= invCounts;
                for (int i = 0; i < n; i++) xxMean[i] *= invCounts;
                for (int i = 0; i < n; i++)
                    invMass[i] = xxMean[i] - xMean[i] * xMean[i] + 0.01;
                for (int i = 0; i < n; i++)
                    massStd[i] = sqrt(1.0 / invMass[i]);
            }
        }

        /* Radford Neal's adaptation scheme for step size */
        if (wsteps >= adaptStart && wsteps <= adaptEnd) {
            windowCount++;
            if (windowCount == 10) {
                if (acceptCount == 0) {
                    /* If last 10 were rejected, decrease eps by 20% */
                    epsMu += log(0.8);
                } else if (acceptCount > 8) {
                    /* If >8 were accepted, increase eps by 20% */
                    epsMu += log(1.2);
                }
                acceptCount = 0;
                windowCount = 0;
            }
        }
    } while (wsteps < warmup);

    /* Tighten the step size variation during sampling */
    *epsilon = exp(epsMu);
    epsStd = log(2.0);

    /* --------------------------- Sampling steps --------------------------- */
    int sx = 0;
    int nsteps = 0;
    do {
        /* Randomize momentum */
        for (int i = 0; i < n; i++) p[i] = RandomNormal() * massStd[i];
        numeric_t K = 0;
        for (int i = 0; i < n; i++) K += p[i] * p[i] * invMass[i];
        K *= 0.5;
        numeric_t H = U + K;

        /* Integrate system by L leapfrog steps */
        eps = exp(epsStd * RandomNormal() + epsMu);
        for (int i = 0; i < n; i++) xnew[i] = x[i];
        for (int i = 0; i < n; i++) gnew[i] = g[i];
        for (int lx = 0; lx < L; lx++) {
            for (int i = 0; i < n; i++) p[i] -= eps * gnew[i] * 0.5;
            for (int i = 0; i < n; i++) xnew[i] += eps * p[i] * invMass[i];
            grad(params, xnew, gnew, n);
            for (int i = 0; i < n; i++) p[i] -= eps * gnew[i] * 0.5;
        }

        /* Accept or reject */
        numeric_t Unew = hfun(params, xnew, n);
        numeric_t Knew = 0; 
        for (int i = 0; i < n; i++) Knew += p[i] * p[i] * invMass[i];
        Knew *= 0.5;
        numeric_t Hnew = Unew + Knew;
        numeric_t dH = Hnew - H;
        if (dH < 0 || ((numeric_t) genrand_real3()) < exp(-dH)) {
            U = Unew;
            for (int i = 0; i < n; i++) x[i] = xnew[i];
            for (int i = 0; i < n; i++) g[i] = gnew[i];
            for (int i = 0; i < n; i++) X[sx * n + i] = xnew[i];
            sx++;
        }
        nsteps++;
    } while (sx < s && nsteps < 20 * s);

    free(x);
    free(xnew);
    free(g);
    free(gnew);
    free(p);
    free(invMass);
    free(xMean);
    free(xxMean);
    free(massStd);

    numeric_t accRate = (numeric_t) sx / (numeric_t) nsteps;
    return accRate;
}

numeric_t RandomNormal() {
    /* Generates a standard normal with zero mean and unit variance by the 
       polar method (en.wikipedia.org/wiki/Marsaglia_polar_method) */
    numeric_t u, v, s;
    do {
        u = (numeric_t) genrand_real3() * 2 - 1;
        v = (numeric_t) genrand_real3() * 2 - 1;
        s = u * u + v * v;
    } while (s >= 1 || s == 0);
    numeric_t mul = sqrt(-2.0 * log(s) / s);
    return u * mul;
}

numeric_t HMCSiteH(void *params, const numeric_t *x, const int n) {
    /* Computes the (unnormalized) negative log posterior of single-site 
       hierarchical model */
    numeric_t f = 0;

    void **d = (void **)params;
    numeric_t *C = (numeric_t *) d[0];
    numeric_t scale = *((numeric_t *) d[1]);
    // numeric_t counts = *((numeric_t *) d[2]);
    numeric_t sigma = exp(x[0]);
    int nCodes = n - 1;

    /* Model distribution              p = exp(h) / sum(exp(h))  */
    numeric_t *p = (numeric_t *) malloc(nCodes * sizeof(numeric_t));
    for (int i = 0; i < nCodes; i++) p[i] = exp(x[i + 1]);
    numeric_t Z = 0;
    for (int i = 0; i < nCodes; i++) Z += p[i];
    numeric_t Zinv = 1.0 / Z;
    for (int i = 0; i < nCodes; i++) p[i] *= Zinv;

    /* Negative log likelihood     -loglk = -sum(C .* log(p)) */
    for (int i = 0; i < nCodes; i++) f -= C[i] * log(p[i]);
    free(p);

    /* Negative log prior */
    numeric_t ssq = 0;
    for (int i = 0; i < nCodes; i++) ssq += x[i + 1] * x[i + 1];
    f += (1.0 / (2.0 * sigma * sigma)) * ssq + nCodes * x[0];

    /* Negative log hyperprior */
    f += -x[0] + log(scale * scale + sigma * sigma);

    return f;
}

void HMCSiteHGrad(void *params, const numeric_t *x, numeric_t *g, const int n) {
    /* Computes the gradient of the (unnormalized) negative log posterior of 
       a single-site hierarchical model */

    void **d = (void **)params;
    numeric_t *C = (numeric_t *) d[0];
    numeric_t scale = *((numeric_t *) d[1]);
    numeric_t counts = *((numeric_t *) d[2]);
    numeric_t sigma = exp(x[0]);
    int nCodes = n - 1;

    /* Model distribution              p = exp(h) / sum(exp(h))  */
    numeric_t *p = (numeric_t *) malloc(nCodes * sizeof(numeric_t));
    for (int i = 0; i < nCodes; i++) p[i] = exp(x[i + 1]);
    numeric_t Z = 0;
    for (int i = 0; i < nCodes; i++) Z += p[i];
    numeric_t Zinv = 1.0 / Z;
    for (int i = 0; i < nCodes; i++) p[i] *= Zinv;

    /* Negative log prior */
    numeric_t invSS = 1.0 / (sigma * sigma);
    numeric_t ssq = 0;
    for (int i = 0; i < nCodes; i++) ssq += x[i + 1] * x[i + 1];
    g[0] = -invSS * ssq + nCodes;
    for (int i = 0; i < nCodes; i++)
        g[i + 1] = -counts * (C[i] / counts - p[i]) + invSS * x[i + 1];

    /* Negative log hyperprior */
    g[0] += -1 + 2 * sigma * sigma / (scale * scale + sigma * sigma);

    free(p);
}

numeric_t HMCSiteHNonCenter(void *params, const numeric_t *x, const int n) {
    /* Computes the (unnormalized) negative log posterior of single-site 
       hierarchical model */
    numeric_t f = 0;

    void **d = (void **)params;
    numeric_t *C = (numeric_t *) d[0];
    numeric_t scale = *((numeric_t *) d[1]);
    // numeric_t counts = *((numeric_t *) d[2]);
    numeric_t sigma = exp(x[0]);
    int nCodes = n - 1;

    /* Model distribution              p = exp(h) / sum(exp(h))  */
    numeric_t *p = (numeric_t *) malloc(nCodes * sizeof(numeric_t));
    for (int i = 0; i < nCodes; i++) p[i] = exp(sigma * x[i + 1]);
    numeric_t Z = 0;
    for (int i = 0; i < nCodes; i++) Z += p[i];
    numeric_t Zinv = 1.0 / Z;
    for (int i = 0; i < nCodes; i++) p[i] *= Zinv;

    /* Negative log likelihood     -loglk = -sum(C .* log(p)) */
    for (int i = 0; i < nCodes; i++) f -= C[i] * log(p[i]);
    free(p);

    /* Negative log prior */
    numeric_t ssq = 0;
    for (int i = 0; i < nCodes; i++) ssq += x[i + 1] * x[i + 1];
    f += ssq / 2.0;

    /* Negative log hyperprior */
    f += -x[0] + log(scale * scale + sigma * sigma);

    return f;
}

void HMCSiteHGradNonCenter(void *params, const numeric_t *x, numeric_t *g, 
    const int n) {
    /* Computes the gradient of the (unnormalized) negative log posterior of 
       a single-site hierarchical model */

    void **d = (void **)params;
    numeric_t *C = (numeric_t *) d[0];
    numeric_t scale = *((numeric_t *) d[1]);
    numeric_t counts = *((numeric_t *) d[2]);
    numeric_t sigma = exp(x[0]);
    int nCodes = n - 1;

    /* Model distribution              p = exp(h) / sum(exp(h))  */
    numeric_t *p = (numeric_t *) malloc(nCodes * sizeof(numeric_t));
    for (int i = 0; i < nCodes; i++) p[i] = exp(sigma * x[i + 1]);
    numeric_t Z = 0;
    for (int i = 0; i < nCodes; i++) Z += p[i];
    numeric_t Zinv = 1.0 / Z;
    for (int i = 0; i < nCodes; i++) p[i] *= Zinv;

    /* Negative log likelihood + log prior */
    numeric_t avH = 0;
    for (int i = 0; i < nCodes; i++) avH += p[i] * x[i + 1];
    numeric_t gradSum = 0;
    for (int i = 0; i < nCodes; i++) gradSum += C[i] * (x[i + 1] - avH);
    g[0] = -sigma * gradSum;

    for (int i = 0; i < nCodes; i++)
        g[i + 1] = sigma * counts * (p[i] - C[i] / counts) + x[i + 1];

    /* Negative log hyperprior */
    g[0] += -1 + 2 * sigma * sigma / (scale * scale + sigma * sigma);

    free(p);
}

numeric_t QuickSelect(numeric_t *A, int len, int k) {
/* Returns the kth smallest element of A, for use in quantile estimates
   The xth percentile can be estimated by k = floor(x/100 * len) */
    int pivot = 0;
    /* Swap from left end until the pivot is smaller */
    for (int i = 0; i < len - 1; i++) {
        if (A[i] > A[len - 1]) continue;
        numeric_t temp = A[i];
        A[i] = A[pivot];
        A[pivot] = temp;
        pivot++;
    }

    /* Swap the pivot to right end */
    numeric_t temp = A[len - 1];
    A[len - 1] = A[pivot];
    A[pivot] = temp;

    if (k == pivot) {
        return A[pivot];
    } else if (pivot > k) {
        return QuickSelect(A, pivot, k);
    } else {
        return QuickSelect(&(A[pivot]), len - pivot, k - pivot);
    }
}

static lbfgsfloatval_t PLMNegLogPosterior(void *instance,
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n,
    const lbfgsfloatval_t step) {
    /* Compute the the negative log posterior, which is the negative 
       penalized log-(pseudo)likelihood and the objective for MAP inference
    */
    void **d = (void **)instance;
    alignment_t *ali = (alignment_t *) d[0];
    options_t *options = (options_t *) d[1];
    numeric_t *lambdas = (numeric_t *) d[2];

    /* Initialize log-likelihood and gradient */
    lbfgsfloatval_t fx = 0.0;
    for (int i = 0; i < ali->nParams; i++) g[i] = 0;

    /* Profiling code START */
    #if defined(PROFILE_TIMES)
        struct timeval tic;
        gettimeofday(&tic, NULL);
    #endif

    /* Negative log-pseudolikelihood */
    #pragma omp parallel for
    for (int i = 0; i < ali->nSites; i++) {
        numeric_t *H = (numeric_t *) malloc(ali->nCodes * sizeof(numeric_t));
        numeric_t *P = (numeric_t *) malloc(ali->nCodes * sizeof(numeric_t));

        numeric_t siteFx = 0.0;
        /* Reshape site parameters and gradient into local blocks */
        numeric_t *Xi = (numeric_t *) malloc(ali->nCodes * ali->nCodes
            * ali->nSites * sizeof(numeric_t));
        for (int j = 0; j < i; j++)
            for (int a = 0; a < ali->nCodes; a++)
                for (int b = 0; b < ali->nCodes; b++)
                    siteE(j, a, b) = xEij(i, j, a, b);
        for (int j = i + 1; j < ali->nSites; j++)
            for (int a = 0; a < ali->nCodes; a++)
                for (int b = 0; b < ali->nCodes; b++)
                    siteE(j, a, b) = xEij(i, j, a, b);
        for (int a = 0; a < ali->nCodes; a++) siteH(i, a) = xHi(i, a);

        numeric_t *Di = (numeric_t *) malloc(ali->nCodes * ali->nCodes
        * ali->nSites * sizeof(numeric_t));
        for (int d = 0; d < ali->nCodes * ali->nCodes * ali->nSites; d++)
            Di[d] = 0.0;

        /* Site negative conditional log likelihoods */
        for (int s = 0; s < ali->nSeqs; s++) {
            /* Compute potentials */
            for (int a = 0; a < ali->nCodes; a++) H[a] = siteH(i, a);
            for (int j = 0; j < i; j++)
                for (int a = 0; a < ali->nCodes; a++)
                    H[a] += siteE(j, a, seq(s, j));
            for (int j = i + 1; j < ali->nSites; j++)
                for (int a = 0; a < ali->nCodes; a++)
                    H[a] += siteE(j, a, seq(s, j));

            /* Conditional distribution given sequence background */
            numeric_t scale = H[0];
            for (int a = 1; a < ali->nCodes; a++)
                scale = (scale >= H[a] ? scale : H[a]);
            for (int a = 0; a < ali->nCodes; a++) P[a] = exp(H[a] - scale);
            numeric_t Z = 0;
            for (int a = 0; a < ali->nCodes; a++) Z += P[a];
            numeric_t Zinv = 1.0 / Z;
            for (int a = 0; a < ali->nCodes; a++) P[a] *= Zinv;


            /* Log-likelihood contributions are scaled by sequence weight */
            numeric_t w = ali->weights[s];	
            siteFx -= w * log(P[seq(s, i)]);

            /* Field gradient */
            siteDH(i, seq(s, i)) -= w;
            for (int a = 0; a < ali->nCodes; a++)
                siteDH(i, a) -= -w * P[a];

            /* Couplings gradient */
            int ix = seq(s, i);
            for (int j = 0; j < i; j++)
                siteDE(j, ix, seq(s, j)) -= w;
            for (int j = i + 1; j < ali->nSites; j++)
                siteDE(j, ix, seq(s, j)) -= w;
            for (int j = 0; j < i; j++)
                for (int a = 0; a < ali->nCodes; a++)
                    siteDE(j, a, seq(s, j)) -= -w * P[a];
            for (int j = i + 1; j < ali->nSites; j++)
                for (int a = 0; a < ali->nCodes; a++)
                    siteDE(j, a, seq(s, j)) -= -w * P[a];
        }

        /* Contribute local loglk and gradient to global */
        #pragma omp critical
        {
        fx += siteFx;
        for (int j = 0; j < i; j++)
            for (int a = 0; a < ali->nCodes; a++)
                for (int b = 0; b < ali->nCodes; b++)
                    dEij(i, j, a, b) += siteDE(j, a, b);
        for (int j = i + 1; j < ali->nSites; j++)
            for (int a = 0; a < ali->nCodes; a++)
                for (int b = 0; b < ali->nCodes; b++)
                    dEij(i, j, a, b) += siteDE(j, a, b);
        for (int a = 0; a < ali->nCodes; a++) dHi(i, a) += siteDH(i, a);
        free(Xi);
        free(Di);
        }

        free(H);
        free(P);
    }

    /* Profiling code STOP */
    #if defined(PROFILE_TIMES)
        struct timeval toc;
        gettimeofday(&toc, NULL);
        /* Subtraction routines from FastTree */
        if (toc.tv_usec < tic.tv_usec) {
            int nsec = (tic.tv_usec - toc.tv_usec) / 1000000 + 1;
            tic.tv_usec -= 1000000 * nsec;
            tic.tv_sec += nsec;
        }
        if (toc.tv_usec - tic.tv_usec > 1000000) {
            int nsec = (toc.tv_usec - tic.tv_usec) / 1000000;
            tic.tv_usec += 1000000 * nsec;
            tic.tv_sec -= nsec;
        }
        int usec = (int) (toc.tv_usec - tic.tv_usec);
        int sec = (int) (toc.tv_sec - tic.tv_sec);
        fprintf(stderr, " %i.%2i seconds\n", sec, usec/ 10000);
    #endif

    ali->negLogLk = fx;

    /* Gaussian priors */
    for (int i = 0; i < ali->nSites; i++)
        for (int ai = 0; ai < ali->nCodes; ai++) {
            dHi(i, ai) += lambdaHi(i) * 2.0 * xHi(i, ai);
            fx += lambdaHi(i) * xHi(i, ai) * xHi(i, ai);
        }

    for (int i = 0; i < ali->nSites-1; i++)
        for (int j = i + 1; j < ali->nSites; j++)
            for (int ai = 0; ai < ali->nCodes; ai++)
                for (int aj = 0; aj < ali->nCodes; aj++) {
                    dEij(i, j, ai, aj) += lambdaEij(i, j)
                        * 2.0 * xEij(i, j, ai, aj);
                    fx += lambdaEij(i, j)
                        * xEij(i, j, ai, aj) * xEij(i, j, ai, aj);
                }

    PostCondition(x, g, ali, options);
    return fx;
}

static lbfgsfloatval_t PLMNegLogPosteriorGapReduce(void *instance,
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n,
    const lbfgsfloatval_t step) {
    /* Compute the the negative log posterior, which is the negative 
       penalized log-(pseudo)likelihood and the objective for MAP inference
    */
    void **d = (void **)instance;
    alignment_t *ali = (alignment_t *) d[0];
    options_t *options = (options_t *) d[1];
    numeric_t *lambdas = (numeric_t *) d[2];

    /* Initialize log-likelihood and gradient */
    lbfgsfloatval_t fx = 0.0;
    for (int i = 0; i < ali->nParams; i++) g[i] = 0;

    /* Profiling code START */
    #if defined(PROFILE_TIMES)
        struct timeval tic;
        gettimeofday(&tic, NULL);
    #endif

    /* Negative log-pseudolikelihood */
    #pragma omp parallel for
    for (int i = 0; i < ali->nSites; i++) {
        numeric_t *H = (numeric_t *) malloc(ali->nCodes * sizeof(numeric_t));
        numeric_t *P = (numeric_t *) malloc(ali->nCodes * sizeof(numeric_t));

        numeric_t siteFx = 0.0;
        /* Reshape site parameters and gradient into local blocks */
        numeric_t *Xi = (numeric_t *) malloc(ali->nCodes * ali->nCodes
            * ali->nSites * sizeof(numeric_t));
        for (int j = 0; j < i; j++)
            for (int a = 0; a < ali->nCodes; a++)
                for (int b = 0; b < ali->nCodes; b++)
                    siteE(j, a, b) = xEij(i, j, a, b);
        for (int j = i + 1; j < ali->nSites; j++)
            for (int a = 0; a < ali->nCodes; a++)
                for (int b = 0; b < ali->nCodes; b++)
                    siteE(j, a, b) = xEij(i, j, a, b);
        for (int a = 0; a < ali->nCodes; a++) siteH(i, a) = xHi(i, a);

        numeric_t *Di = (numeric_t *) malloc(ali->nCodes * ali->nCodes
        * ali->nSites * sizeof(numeric_t));
        for (int d = 0; d < ali->nCodes * ali->nCodes * ali->nSites; d++)
            Di[d] = 0.0;

        /* Site negative conditional log likelihoods */
        for (int s = 0; s < ali->nSeqs; s++) {
            /* Only ungapped sites are considered in the model */
            if (seq(s, i) >= 0) {
                /* Compute potentials */
                for (int a = 0; a < ali->nCodes; a++) H[a] = siteH(i, a);
                for (int j = 0; j < i; j++)
                    for (int a = 0; a < ali->nCodes; a++)
                        if (seq(s, j) >= 0)
                            H[a] += siteE(j, a, seq(s, j));
                for (int j = i + 1; j < ali->nSites; j++)
                    for (int a = 0; a < ali->nCodes; a++)
                        if (seq(s, j) >= 0)
                            H[a] += siteE(j, a, seq(s, j));

                /* Conditional distribution given sequence background */
                numeric_t scale = H[0];
                for (int a = 1; a < ali->nCodes; a++)
                    scale = (scale >= H[a] ? scale : H[a]);
                for (int a = 0; a < ali->nCodes; a++) P[a] = exp(H[a] - scale);
                numeric_t Z = 0;
                for (int a = 0; a < ali->nCodes; a++) Z += P[a];
                numeric_t Zinv = 1.0 / Z;
                for (int a = 0; a < ali->nCodes; a++) P[a] *= Zinv;


                /* Log-likelihood contributions are scaled by sequence weight */
                numeric_t w = ali->weights[s];  
                siteFx -= w * log(P[seq(s, i)]);

                /* Field gradient */
                siteDH(i, seq(s, i)) -= w;
                for (int a = 0; a < ali->nCodes; a++)
                    siteDH(i, a) -= -w * P[a];

                /* Couplings gradient */
                int ix = seq(s, i);
                for (int j = 0; j < i; j++)
                    if (seq(s, j) >= 0)
                        siteDE(j, ix, seq(s, j)) -= w;
                for (int j = i + 1; j < ali->nSites; j++)
                    if (seq(s, j) >= 0)
                        siteDE(j, ix, seq(s, j)) -= w;
                for (int j = 0; j < i; j++)
                    if (seq(s, j) >= 0)
                        for (int a = 0; a < ali->nCodes; a++)
                            siteDE(j, a, seq(s, j)) -= -w * P[a];
                for (int j = i + 1; j < ali->nSites; j++)
                    if (seq(s, j) >= 0)
                        for (int a = 0; a < ali->nCodes; a++)
                            siteDE(j, a, seq(s, j)) -= -w * P[a];
            }
        }

        /* Contribute local loglk and gradient to global */
        #pragma omp critical
        {
        fx += siteFx;
        for (int j = 0; j < i; j++)
            for (int a = 0; a < ali->nCodes; a++)
                for (int b = 0; b < ali->nCodes; b++)
                    dEij(i, j, a, b) += siteDE(j, a, b);
        for (int j = i + 1; j < ali->nSites; j++)
            for (int a = 0; a < ali->nCodes; a++)
                for (int b = 0; b < ali->nCodes; b++)
                    dEij(i, j, a, b) += siteDE(j, a, b);
        for (int a = 0; a < ali->nCodes; a++) dHi(i, a) += siteDH(i, a);
        free(Xi);
        free(Di);
        }

        free(H);
        free(P);
    }

    /* Profiling code STOP */
    #if defined(PROFILE_TIMES)
        struct timeval toc;
        gettimeofday(&toc, NULL);
        /* Subtraction routines from FastTree */
        if (toc.tv_usec < tic.tv_usec) {
            int nsec = (tic.tv_usec - toc.tv_usec) / 1000000 + 1;
            tic.tv_usec -= 1000000 * nsec;
            tic.tv_sec += nsec;
        }
        if (toc.tv_usec - tic.tv_usec > 1000000) {
            int nsec = (toc.tv_usec - tic.tv_usec) / 1000000;
            tic.tv_usec += 1000000 * nsec;
            tic.tv_sec -= nsec;
        }
        int usec = (int) (toc.tv_usec - tic.tv_usec);
        int sec = (int) (toc.tv_sec - tic.tv_sec);
        fprintf(stderr, " %i.%2i seconds\n", sec, usec/ 10000);
    #endif

    ali->negLogLk = fx;

    /* Gaussian priors */
    for (int i = 0; i < ali->nSites; i++)
        for (int ai = 0; ai < ali->nCodes; ai++) {
            dHi(i, ai) += lambdaHi(i) * 2.0 * xHi(i, ai);
            fx += lambdaHi(i) * xHi(i, ai) * xHi(i, ai);
        }

    for (int i = 0; i < ali->nSites-1; i++)
        for (int j = i + 1; j < ali->nSites; j++)
            for (int ai = 0; ai < ali->nCodes; ai++)
                for (int aj = 0; aj < ali->nCodes; aj++) {
                    dEij(i, j, ai, aj) += lambdaEij(i, j)
                        * 2.0 * xEij(i, j, ai, aj);
                    fx += lambdaEij(i, j)
                        * xEij(i, j, ai, aj) * xEij(i, j, ai, aj);
                }

    PostCondition(x, g, ali, options);
    return fx;
}

static lbfgsfloatval_t PLMNegLogPosteriorBlock(void *instance,
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n,
    const lbfgsfloatval_t step) {
    /* Compute the the negative log posterior, which is the negative 
       penalized log-(pseudo)likelihood and the objective for MAP inference
    */
    void **d = (void **)instance;
    alignment_t *ali = (alignment_t *) d[0];
    options_t *options = (options_t *) d[1];
    numeric_t *lambdas = (numeric_t *) d[2];

    /* Initialize log-likelihood and gradient */
    lbfgsfloatval_t fx = 0.0;
    for (int i = 0; i < ali->nParams; i++) g[i] = 0;

    /* Profiling code START */
    #if defined(PROFILE_TIMES)
        struct timeval tic;
        gettimeofday(&tic, NULL);
    #endif

    /* Block fields hi */
    numeric_t *hi = (numeric_t *)
        malloc(ali->nSites * ali->nCodes * sizeof(numeric_t));
    numeric_t *gHi = (numeric_t *)
        malloc(ali->nSites * ali->nCodes * sizeof(numeric_t));
    for (int i = 0; i < ali->nSites; i++)
        for (int ai = 0; ai < ali->nCodes; ai++) Hi(i, ai) = xHi(i, ai);
    for (int i = 0; i < ali->nSites * ali->nCodes; i++) gHi[i] = 0;

    /* Block couplings eij */
    numeric_t *eij = (numeric_t *) malloc(ali->nSites * ali->nSites
        * ali->nCodes * ali->nCodes * sizeof(numeric_t));
    numeric_t *gEij = (numeric_t *) malloc(ali->nSites * ali->nSites
        * ali->nCodes * ali->nCodes * sizeof(numeric_t));
    for (int i = 0; i < ali->nSites * ali->nSites * ali->nCodes * ali->nCodes;
        i++) eij[i] = 0.0;
    for (int i = 0; i < ali->nSites * ali->nSites * ali->nCodes * ali->nCodes;
        i++) gEij[i] = 0.0;
    for (int i = 0; i < ali->nSites - 1; i++)
        for (int j = i + 1; j < ali->nSites; j++)
            for (int ai = 0; ai < ali->nCodes; ai++)
                for (int aj = 0; aj < ali->nCodes; aj++)
                    Eij(j, aj, i, ai) = Eij(i, ai, j, aj) = xEij(i, j, ai, aj);


    /* Negative log-pseudolikelihood */
    for (int s = 0; s < ali->nSeqs; s++) {
        /* Form potential for conditional log likelihoods at every site */
        numeric_t *H = (numeric_t *)
            malloc(ali->nCodes * ali->nSites * sizeof(numeric_t));
        numeric_t *Z = (numeric_t *) malloc(ali->nSites * sizeof(numeric_t));

        /* Initialize potentials with fields */
        // memcpy(H, hi, ali->nSites * ali->nCodes * sizeof(numeric_t));
        for(int jx = 0; jx < ali->nSites * ali->nCodes; jx++) H[jx] = hi[jx];

        /* Contribute coupling block due to i, ai */
        for (int i = 0; i < ali->nSites; i++) {
            const letter_t ai = seq(s, i);
            const numeric_t *jB = &(Eij(i, ai, 0, 0));
            for(int jx = 0; jx < ali->nSites * ali->nCodes; jx++)
                H[jx] += jB[jx];
        }

        /* Conditional log likelihoods */
        for (int i = 0; i < ali->nSites * ali->nCodes; i++) H[i] = exp(H[i]);
        for (int i = 0; i < ali->nSites; i++) Z[i] = 0;
        for (int i = 0; i < ali->nSites; i++)
            for (int ai = 0; ai < ali->nSites; ai++) Z[i] += Hp(i, ai);
        for (int i = 0; i < ali->nSites; i++)
            for (int ai = 0; ai < ali->nSites; ai++) Hp(i, ai) /= Z[i];

        numeric_t seqFx = 0;
        for (int i = 0; i < ali->nSites; i++)
            seqFx -= ali->weights[s] * log(Hp(i, seq(s, i)));

        for(int jx = 0; jx < ali->nSites * ali->nCodes; jx++)
            H[jx] *= -ali->weights[s];

        for (int i = 0; i < ali->nSites; i++)
            gHi(i, seq(s, i)) -= ali->weights[s];
        for(int jx = 0; jx < ali->nSites * ali->nCodes; jx++) gHi[jx] -= H[jx];

        for (int i = 0; i < ali->nSites - 1; i++)
            for (int j = i; j < ali->nSites; j++)
                gEij(i, seq(s, i), j, seq(s, j)) -= ali->weights[s];

        for (int i = 0; i < ali->nSites; i++) {
            const letter_t ai = seq(s, i);
            numeric_t *jgBlock = &(gEij(i, ai, 0, 0));
            for (int jx = 0; jx < ali->nSites * ali->nCodes; jx++)
                jgBlock[jx] -= H[jx];
        }

        free(H);
        free(Z);
        fx += seqFx;
    }

    for (int i = 0; i < ali->nSites; i++)
        for (int ai = 0; ai < ali->nCodes; ai++)
            dHi(i, ai) += gHi(i, ai);

    for (int i = 0; i < ali->nSites - 1; i++)
        for (int j = i + 1; j < ali->nSites; j++)
            for (int ai = 0; ai < ali->nCodes; ai++)
                for (int aj = 0; aj < ali->nCodes; aj++)
                    dEij(i, j, ai, aj) += gEij(j, aj, i, ai) + gEij(i, ai, j, aj);
    free(hi);
    free(gHi);
    free(eij);
    free(gEij);

    /* Profiling code STOP */
    #if defined(PROFILE_TIMES)
        struct timeval toc;
        gettimeofday(&toc, NULL);
        /* Subtraction routines from FastTree */
        if (toc.tv_usec < tic.tv_usec) {
            int nsec = (tic.tv_usec - toc.tv_usec) / 1000000 + 1;
            tic.tv_usec -= 1000000 * nsec;
            tic.tv_sec += nsec;
        }
        if (toc.tv_usec - tic.tv_usec > 1000000) {
            int nsec = (toc.tv_usec - tic.tv_usec) / 1000000;
            tic.tv_usec += 1000000 * nsec;
            tic.tv_sec -= nsec;
        }
        int usec = (int) (toc.tv_usec - tic.tv_usec);
        int sec = (int) (toc.tv_sec - tic.tv_sec);
        fprintf(stderr, " %i.%2i seconds\n", sec, usec/ 10000);
    #endif

    ali->negLogLk = fx;

    /* Gaussian priors */
    for (int i = 0; i < ali->nSites; i++)
        for (int ai = 0; ai < ali->nCodes; ai++) {
            dHi(i, ai) += lambdaHi(i) * 2.0 * xHi(i, ai);
            fx += lambdaHi(i) * xHi(i, ai) * xHi(i, ai);
        }

    for (int i = 0; i < ali->nSites-1; i++)
        for (int j = i + 1; j < ali->nSites; j++)
            for (int ai = 0; ai < ali->nCodes; ai++)
                for (int aj = 0; aj < ali->nCodes; aj++) {
                    dEij(i, j, ai, aj) += lambdaEij(i, j)
                        * 2.0 * xEij(i, j, ai, aj);
                    fx += lambdaEij(i, j)
                        * xEij(i, j, ai, aj) * xEij(i, j, ai, aj);
                }

    PostCondition(x, g, ali, options);
    return fx;
}

static lbfgsfloatval_t PLMNegLogPosteriorDO(void *instance,
    const lbfgsfloatval_t *x, lbfgsfloatval_t *g, const int n,
    const lbfgsfloatval_t step) {
    /* Compute the the negative log posterior, which is the negative 
       penalized log-(pseudo)likelihood and the objective for MAP inference
    */
    void **d = (void **)instance;
    alignment_t *ali = (alignment_t *) d[0];
    options_t *options = (options_t *) d[1];
    numeric_t *lambdas = (numeric_t *) d[2];

    /* Initialize log-likelihood and gradient */
    lbfgsfloatval_t fx = 0.0;
    for (int i = 0; i < ali->nParams; i++) g[i] = 0;

    /* Profiling code START */
    #if defined(PROFILE_TIMES)
        struct timeval tic;
        gettimeofday(&tic, NULL);
    #endif

    numeric_t *H = (numeric_t *) malloc(ali->nCodes * sizeof(numeric_t));
    numeric_t *P = (numeric_t *) malloc(ali->nCodes * sizeof(numeric_t));
    int *drop_mask = (int *) malloc(ali->nParams * sizeof(int));
    for (int s = 0; s < ali->nSeqs; s++) {
        /* Generate random bit mask over parameters */
        for (int p = 0; p < ali->nParams; p ++)
            drop_mask[p] = (int) rand() % 2;

        /* Pseudolikelihood objective */
        for (int i = 0; i < ali->nSites; i++) {
            for (int a = 0; a < ali->nCodes; a++) H[a] = bitHi(i, a)
                                               * xHi(i, a);
            for (int a = 0; a < ali->nCodes; a++)
                for (int j = 0; j < i; j++)
                    H[a] += bitEij(i, j, a, seq(s, j))
                            * xEij(i, j, a, seq(s, j));
            for (int a = 0; a < ali->nCodes; a++)
                for (int j = i + 1; j < ali->nSites; j++)
                    H[a] += bitEij(i, j, a, seq(s, j))
                            * xEij(i, j, a, seq(s, j));

            /* Compute distribution from potential */
            for (int a = 0; a < ali->nCodes; a++) P[a] = exp(H[a]);
            numeric_t Z = 0;
            for (int a = 0; a < ali->nCodes; a++) Z += P[a];
            numeric_t Zinv = 1.0 / Z;
            for (int a = 0; a < ali->nCodes; a++) P[a] *= Zinv;

            /* Log-likelihood contributions */
            fx -= ali->weights[s] * log(P[seq(s, i)]);

            /* Field gradient */
            dHi(i, seq(s, i)) -= bitHi(i, seq(s, i)) * ali->weights[s];
            for (int a = 0; a < ali->nCodes; a++)
                dHi(i, a) -= -bitHi(i, a) * ali->weights[s] * P[a];

            /* Couplings gradient */
            for (int j = 0; j < i; j++)
                dEij(i, j, seq(s, i), seq(s, j)) -=
                    bitEij(i, j, seq(s, i), seq(s, j)) * ali->weights[s];
            for (int j = i + 1; j < ali->nSites; j++)
                dEij(i, j, seq(s, i), seq(s, j)) -=
                    bitEij(i, j, seq(s, i), seq(s, j)) * ali->weights[s];

            for (int j = 0; j < i; j++)
                for (int a = 0; a < ali->nCodes; a++)
                    dEij(i, j, a, seq(s, j)) -=
                        -bitEij(i, j, a, seq(s, j)) * ali->weights[s] * P[a];
            for (int j = i + 1; j < ali->nSites; j++)
                for (int a = 0; a < ali->nCodes; a++)
                    dEij(i, j, a, seq(s, j)) -=
                        -bitEij(i, j, a, seq(s, j)) * ali->weights[s] * P[a];
        }
    }
    free(H);
    free(P);
    free(drop_mask);

    /* Profiling code STOP */
    #if defined(PROFILE_TIMES)
        struct timeval toc;
        gettimeofday(&toc, NULL);
        /* Subtraction routines from FastTree */
        if (toc.tv_usec < tic.tv_usec) {
            int nsec = (tic.tv_usec - toc.tv_usec) / 1000000 + 1;
            tic.tv_usec -= 1000000 * nsec;
            tic.tv_sec += nsec;
        }
        if (toc.tv_usec - tic.tv_usec > 1000000) {
            int nsec = (toc.tv_usec - tic.tv_usec) / 1000000;
            tic.tv_usec += 1000000 * nsec;
            tic.tv_sec -= nsec;
        }
        int usec = (int) (toc.tv_usec - tic.tv_usec);
        int sec = (int) (toc.tv_sec - tic.tv_sec);
        fprintf(stderr, " %i.%2.3i seconds\n", sec, usec/ 10000);
    #endif

    ali->negLogLk = fx;

    /* Gaussian priors */
    for (int i = 0; i < ali->nSites; i++)
        for (int ai = 0; ai < ali->nCodes; ai++) {
            dHi(i, ai) += lambdaHi(i) * 2.0 * xHi(i, ai);
            fx += lambdaHi(i) * xHi(i, ai) * xHi(i, ai);
        }

    for (int i = 0; i < ali->nSites-1; i++)
        for (int j = i + 1; j < ali->nSites; j++)
            for (int ai = 0; ai < ali->nCodes; ai++)
                for (int aj = 0; aj < ali->nCodes; aj++) {
                    dEij(i, j, ai, aj) += lambdaEij(i, j)
                        * 2.0 * xEij(i, j, ai, aj);
                    fx += lambdaEij(i, j)
                        * xEij(i, j, ai, aj) * xEij(i, j, ai, aj);
                }

    PostCondition(x, g, ali, options);
    return fx;
}

static int ReportProgresslBFGS(void *instance, const lbfgsfloatval_t *x,
    const lbfgsfloatval_t *g, const lbfgsfloatval_t fx,
    const lbfgsfloatval_t xnorm, const lbfgsfloatval_t gnorm,
    const lbfgsfloatval_t step, int n, int k, int ls) {
    void **d = (void **)instance;
    alignment_t *ali = (alignment_t *)d[0];

    /* Compute norms of relevant parameters */
    lbfgsfloatval_t hNorm = 0.0, eNorm = 0.0, hGNorm = 0.0, eGNorm = 0.0;
    for (int i = 0; i < ali->nSites * ali->nCodes; i++)
        hNorm += x[i]*x[i];
    for (int i = 0; i < ali->nSites * ali->nCodes; i++)
        hGNorm += g[i]*g[i];
    for (int i = ali->nSites * ali->nCodes; i < ali->nParams; i++)
        eNorm += x[i]*x[i];
    for (int i = ali->nSites * ali->nCodes; i < ali->nParams; i++)
        eGNorm += g[i]*g[i];
    hNorm = sqrt(hNorm);
    hGNorm = sqrt(hGNorm);
    eNorm = sqrt(eNorm);
    eGNorm = sqrt(eGNorm);

    /* Retrieve elapsed time */
    static struct timeval now;
    gettimeofday(&now, NULL);
    if (now.tv_usec < ali->start.tv_usec) {
        int nsec = (ali->start.tv_usec - now.tv_usec) / 1000000 + 1;
        ali->start.tv_usec -= 1000000 * nsec;
        ali->start.tv_sec += nsec;
    }
    if (now.tv_usec - ali->start.tv_usec > 1000000) {
        int nsec = (now.tv_usec - ali->start.tv_usec) / 1000000;
        ali->start.tv_usec += 1000000 * nsec;
        ali->start.tv_sec -= nsec;
    }
    numeric_t elapsed = (numeric_t) (now.tv_sec - ali->start.tv_sec)
                      + ((numeric_t) (now.tv_usec - ali->start.tv_usec)) / 1E6;

    if (k == 1) fprintf(stderr,
        "iter\ttime\tcond\tfx\t-loglk"
        "\t||h||\t||e||\n");
    fprintf(stderr, "%d\t%.1f\t%.2f\t%.1f\t%.1f\t%.1f\t%.1f\n",
        k, elapsed, gnorm / xnorm, fx, ali->negLogLk, hNorm, eNorm);
    return 0;
}

void PreCondition(const lbfgsfloatval_t *x, lbfgsfloatval_t *g, alignment_t *ali, options_t *options) {
    /* Currently empty */
}

void PostCondition(const lbfgsfloatval_t *x, lbfgsfloatval_t *g, alignment_t *ali, options_t *options) {
    if (options->zeroAPC == 1)
        for (int i = 0; i < ali->nSites; i++)
            for (int ai = 0; ai < ali->nCodes; ai++)
                dHi(i, ai) = 0.0;
}

void ZeroAPCPriors(alignment_t *ali, options_t *options, numeric_t *lambdas,
    lbfgsfloatval_t *x) {
    /* Compute the variances of the couplings for each pair */
    for (int i = 0; i < ali->nSites - 1; i++)
        for (int j = i + 1; j < ali->nSites; j++) {
            /* Mean(eij) over ai, aj */
            numeric_t mean = 0.0;
            for (int ai = 0; ai < ali->nCodes; ai++)
                for (int aj = 0; aj < ali->nCodes; aj++)
                    mean += xEij(i, j, ai, aj);
            mean *= 1.0 / ((numeric_t) ali->nCodes * ali->nCodes);

            /* Var(eij) over ai, aj */
            numeric_t ssq = 0.0;
            for (int ai = 0; ai < ali->nCodes; ai++)
                for (int aj = 0; aj < ali->nCodes; aj++)
                    ssq += (xEij(i, j, ai, aj) - mean)
                         * (xEij(i, j, ai, aj) - mean);
            /* Use N rather than N-1 since N has better MSE */
            numeric_t var = ssq / ((numeric_t) (ali->nCodes * ali->nCodes));
            lambdaEij(i, j) = var;
        }

    /* Determine the site-wise statistics of the variances */
    numeric_t nPairs =  ((numeric_t) ((ali->nSites) * (ali->nSites - 1))) / 2.0;
    numeric_t V_avg = 0.0;
    numeric_t *V_pos_avg = (numeric_t *) malloc(ali->nSites * sizeof(numeric_t));
    for (int i = 0; i < ali->nSites; i++) {
        V_pos_avg[i] = 0.0;
    }
    for (int i = 0; i < ali->nSites - 1; i++) {
        for (int j = i + 1; j < ali->nSites; j++) {
            V_pos_avg[i] += lambdaEij(i, j) / (numeric_t) (ali->nSites - 1);
            V_pos_avg[j] += lambdaEij(i, j) / (numeric_t) (ali->nSites - 1);
            V_avg += lambdaEij(i, j) / nPairs;
        }
    }

    /* Remove the first component of the variances */
    for (int i = 0; i < ali->nSites - 1; i++)
        for (int j = i + 1; j < ali->nSites; j++)
            lambdaEij(i, j) =
                lambdaEij(i, j) - V_pos_avg[i] * V_pos_avg[j] / V_avg;

    /* Transform and truncate variances into lambda hyperparameters */
    numeric_t pcount = 0.0;
    numeric_t psum = 0.0;
    numeric_t inbounds = 0;
    numeric_t min = LAMBDA_J_MAX;
    numeric_t max = LAMBDA_J_MIN;
    for (int i = 0; i < ali->nSites - 1; i++) {
        for (int j = i + 1; j < ali->nSites; j++) {
            /* Lambda coefficients are 1/2 the inverse variance */
            if (lambdaEij(i, j) > 0) {
                lambdaEij(i, j) = 1.0 / (2.0 * lambdaEij(i, j));
                psum += lambdaEij(i, j);
                pcount += 1.0;
            } else {
                lambdaEij(i, j) = LAMBDA_J_MAX + 1.0;
            }

            /* Truncate lambda for numerical stability */
            if (lambdaEij(i, j) >= LAMBDA_J_MIN && lambdaEij(i, j) <= LAMBDA_J_MAX)
                inbounds += 1.0 / (numeric_t) ((ali->nSites)*(ali->nSites - 1) / 2.0);
            if (lambdaEij(i, j) < 0 || !isfinite(lambdaEij(i, j)))
                lambdaEij(i, j) = LAMBDA_J_MAX;
            if (lambdaEij(i, j) < LAMBDA_J_MIN) lambdaEij(i, j) = LAMBDA_J_MIN;
            if (lambdaEij(i, j) > LAMBDA_J_MAX) lambdaEij(i, j) = LAMBDA_J_MAX;

            /* Track extremes */
            if (lambdaEij(i, j) > max) max = lambdaEij(i, j);
            if (lambdaEij(i, j) < min) min = lambdaEij(i, j);
        }
    }
    fprintf(stderr, "Raw coupling hyperparameter statistics:\n"
                    "\tMean positive lambda: %f\n"
                    "\tPercent of ij's positive: %f\n"
                    "\tPercent in bounds (%f < L < %f): %f\n",
                    psum / pcount,
                    pcount / nPairs,
                    min, max, inbounds);
}

const char *LBFGSErrorString(int ret) {
    const char *p;
    switch(ret) {
        case LBFGSERR_UNKNOWNERROR:
            p = "UNKNOWNERROR";
            break;
        /** Logic error. */
        case LBFGSERR_LOGICERROR:
            p = "LOGICERROR";
            break;
        /** Insufficient memory. */
        case LBFGSERR_OUTOFMEMORY:
            p = "OUTOFMEMORY";
            break;
        /** The minimization process has been canceled. */
        case LBFGSERR_CANCELED:
            p = "CANCELED";
            break;
        /** Invalid number of variables specified. */
        case LBFGSERR_INVALID_N:
            p = "INVALID_N";
            break;
        /** Invalid number of variables (for SSE) specified. */
        case LBFGSERR_INVALID_N_SSE:
            p = "INVALID_N_SSE";
            break;
        /** The array x must be aligned to 16 (for SSE). */
        case LBFGSERR_INVALID_X_SSE:
            p = "INVALID_X_SSE";
            break;
        /** Invalid parameter lbfgs_parameter_t::epsilon specified. */
        case LBFGSERR_INVALID_EPSILON:
            p = "INVALID_EPSILON";
            break;
        /** Invalid parameter lbfgs_parameter_t::past specified. */
        case LBFGSERR_INVALID_TESTPERIOD:
            p = "INVALID_TESTPERIOD";
            break;
        /** Invalid parameter lbfgs_parameter_t::delta specified. */
        case LBFGSERR_INVALID_DELTA:
            p = "INVALID_DELTA";
            break;
        /** Invalid parameter lbfgs_parameter_t::linesearch specified. */
        case LBFGSERR_INVALID_LINESEARCH:
            p = "INVALID_LINESEARCH";
            break;
        /** Invalid parameter lbfgs_parameter_t::max_step specified. */
        case LBFGSERR_INVALID_MINSTEP:
            p = "INVALID_MINSTEP";
            break;
        /** Invalid parameter lbfgs_parameter_t::max_step specified. */
        case LBFGSERR_INVALID_MAXSTEP:
            p = "INVALID_MAXSTEP";
            break;
        /** Invalid parameter lbfgs_parameter_t::ftol specified. */
        case LBFGSERR_INVALID_FTOL:
            p = "INVALID_FTOL";
            break;
        /** Invalid parameter lbfgs_parameter_t::wolfe specified. */
        case LBFGSERR_INVALID_WOLFE:
            p = "INVALID_WOLFE";
            break;
        /** Invalid parameter lbfgs_parameter_t::gtol specified. */
        case LBFGSERR_INVALID_GTOL:
            p = "INVALID_GTOL";
            break;
        /** Invalid parameter lbfgs_parameter_t::xtol specified. */
        case LBFGSERR_INVALID_XTOL:
            p = "INVALID_XTOL";
            break;
        /** Invalid parameter lbfgs_parameter_t::max_linesearch specified. */
        case LBFGSERR_INVALID_MAXLINESEARCH:
            p = "INVALID_MAXLINESEARCH";
            break;
        /** Invalid parameter lbfgs_parameter_t::orthantwise_c specified. */
        case LBFGSERR_INVALID_ORTHANTWISE:
            p = "INVALID_ORTHANTWISE";
            break;
        /** Invalid parameter lbfgs_parameter_t::orthantwise_start specified. */
        case LBFGSERR_INVALID_ORTHANTWISE_START:
            p = "INVALID_ORTHANTWISE_START";
            break;
        /** Invalid parameter lbfgs_parameter_t::orthantwise_end specified. */
        case LBFGSERR_INVALID_ORTHANTWISE_END:
            p = "ORTHANTWISE_END";
            break;
        /** The line-search step went out of the interval of uncertainty. */
        case LBFGSERR_OUTOFINTERVAL:
            p = "OUTOFINTERVAL";
            break;
        /** A logic error occurred; alternatively: the interval of uncertainty
            became too small. */
        case LBFGSERR_INCORRECT_TMINMAX:
            p = "INCORRECT_TMINMAX";
            break;
        /** A rounding error occurred; alternatively: no line-search step
            satisfies the sufficient decrease and curvature conditions. */
        case LBFGSERR_ROUNDING_ERROR:
            p = "ROUNDING_ERROR";
            break;
        /** The line-search step became smaller than lbfgs_parameter_t::min_step. */
        case LBFGSERR_MINIMUMSTEP:
            p = "MINIMUMSTEP";
            break;
        /** The line-search step became larger than lbfgs_parameter_t::max_step. */
        case LBFGSERR_MAXIMUMSTEP:
            p = "MAXILBFGSERR_MUMSTEP";
            break;
        /** The line-search routine reaches the maximum number of evaluations. */
        case LBFGSERR_MAXIMUMLINESEARCH:
            p = "MAXIMUMLINESEARCH";
            break;
        /** The algorithm routine reaches the maximum number of iterations. */
        case LBFGSERR_MAXIMUMITERATION:
            p = "MAXIMUMITERATION";
            break;
        /** Relative width of the interval of uncertainty is at most
            lbfgs_parameter_t::xtol. */
        case LBFGSERR_WIDTHTOOSMALL:
            p = "WIDTHTOOSMALL";
            break;
        /** A logic error (negative line-search step) occurred. */
        case LBFGSERR_INVALIDPARAMETERS:
            p = "INVALIDPARAMETERS";
            break;
        /** The current search direction increases the objective function value. */
        case LBFGSERR_INCREASEGRADIENT:
            p = "INCREASEGRADIENT";
            break;
        case 0:
            p = "Minimization success";
            break;
        default:
            p = "No detected error";
            break;
    }
    return p;
}